@@ -1,7 +1,15 @@
-# Coldcard interaction script
+"""
+Coldcard
+********
+"""
 
-from typing import Dict, Union
+from typing import (
+    Dict,
+    List,
+    Union,
+)
 
+from ..descriptor import MultisigDescriptor
 from ..hwwclient import HardwareWalletClient
 from ..errors import (
     ActionCanceledError,
@@ -32,17 +40,20 @@ from .ckcc.constants import (
     AF_P2SH,
     AF_P2WSH_P2SH,
 )
-from .ckcc.utils import (
-    str_to_int_path,
-)
-from ..base58 import (
+from .._base58 import (
     get_xpub_fingerprint,
-    xpub_main_2_test,
 )
-from ..serializations import (
+from ..key import (
     ExtendedKey,
+)
+from ..psbt import (
     PSBT,
 )
+from ..common import (
+    AddressType,
+    Chain,
+)
+from functools import wraps
 from hashlib import sha256
 
 import base64
@@ -51,13 +62,20 @@ import io
 import sys
 import time
 import struct
-from binascii import hexlify, a2b_hex, b2a_hex
+
+from binascii import b2a_hex
+from typing import (
+    Any,
+    Callable,
+)
 
 CC_SIMULATOR_SOCK = '/tmp/ckcc-simulator.sock'
 # Using the simulator: https://github.com/Coldcard/firmware/blob/master/unix/README.md
 
-def coldcard_exception(f):
-    def func(*args, **kwargs):
+
+def coldcard_exception(f: Callable[..., Any]) -> Callable[..., Any]:
+    @wraps(f)
+    def func(*args: Any, **kwargs: Any) -> Any:
         try:
             return f(*args, **kwargs)
         except CCProtoError as e:
@@ -71,7 +89,7 @@ def coldcard_exception(f):
 # This class extends the HardwareWalletClient for ColdCard specific things
 class ColdcardClient(HardwareWalletClient):
 
-    def __init__(self, path, password='', expert=False):
+    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
         super(ColdcardClient, self).__init__(path, password, expert)
         # Simulator hard coded pipe socket
         if path == CC_SIMULATOR_SOCK:
@@ -81,32 +99,30 @@ class ColdcardClient(HardwareWalletClient):
             device.open_path(path.encode())
             self.device = ColdcardDevice(dev=device)
 
-    # Must return a dict with the xpub
-    # Retrieves the public key at the specified BIP 32 derivation path
     @coldcard_exception
-    def get_pubkey_at_path(self, path):
+    def get_pubkey_at_path(self, path: str) -> ExtendedKey:
         self.device.check_mitm()
         path = path.replace('h', '\'')
         path = path.replace('H', '\'')
-        xpub = self.device.send_recv(CCProtocolPacker.get_xpub(path), timeout=None)
-        if self.is_testnet:
-            result = {'xpub': xpub_main_2_test(xpub)}
-        else:
-            result = {'xpub': xpub}
-        if self.expert:
-            xpub_obj = ExtendedKey()
-            xpub_obj.deserialize(xpub)
-            result.update(xpub_obj.get_printable_dict())
-        return result
+        xpub_str = self.device.send_recv(CCProtocolPacker.get_xpub(path), timeout=None)
+        xpub = ExtendedKey.deserialize(xpub_str)
+        if self.chain != Chain.MAIN:
+            xpub.version = ExtendedKey.TESTNET_PUBLIC
+        return xpub
 
-    def get_master_fingerprint_hex(self):
+    def get_master_fingerprint(self) -> bytes:
         # quick method to get fingerprint of wallet
-        return hexlify(struct.pack('<I', self.device.master_fingerprint)).decode()
+        return struct.pack('<I', self.device.master_fingerprint)
 
-    # Must return a hex string with the signed transaction
-    # The tx must be in the combined unsigned transaction format
     @coldcard_exception
-    def sign_tx(self, tx):
+    def sign_tx(self, tx: PSBT) -> PSBT:
+        """
+        Sign a transaction with the Coldcard.
+
+        - The Coldcard firmware only supports signing single key and multisig transactions. It cannot sign arbitrary scripts.
+        - Multisigs need to be registered on the device before a transaction spending that multisig will be signed by the device.
+        - Multisigs must use BIP 67. This can be accomplished in Bitcoin Core using the `sortedmulti()` descriptor, available in Bitcoin Core 0.20.
+        """
         self.device.check_mitm()
 
         # Get this devices master key fingerprint
@@ -119,7 +135,7 @@ class ColdcardClient(HardwareWalletClient):
             our_keys = 0
             for key in psbt_in.hd_keypaths.keys():
                 keypath = psbt_in.hd_keypaths[key]
-                if keypath[0] == master_fp and key not in psbt_in.partial_sigs:
+                if keypath.fingerprint == master_fp and key not in psbt_in.partial_sigs:
                     our_keys += 1
             if our_keys > passes:
                 passes = our_keys
@@ -174,10 +190,11 @@ class ColdcardClient(HardwareWalletClient):
 
             tx = PSBT()
             tx.deserialize(base64.b64encode(result).decode())
-        return {'psbt': tx.serialize()}
+
+        return tx
 
     @coldcard_exception
-    def sign_message(self, message: Union[str, bytes], keypath: str) -> Dict[str, str]:
+    def sign_message(self, message: Union[str, bytes], keypath: str) -> str:
         self.device.check_mitm()
         keypath = keypath.replace('h', '\'')
         keypath = keypath.replace('H', '\'')
@@ -206,75 +223,115 @@ class ColdcardClient(HardwareWalletClient):
         _, raw = done
 
         sig = str(base64.b64encode(raw), 'ascii').replace('\n', '')
-        return {"signature": sig}
+        return sig
 
-    # Display address of specified type on the device.
     @coldcard_exception
-    def display_address(self, keypath, p2sh_p2wpkh, bech32, redeem_script=None, descriptor=None):
+    def display_singlesig_address(
+        self,
+        keypath: str,
+        addr_type: AddressType,
+    ) -> str:
         self.device.check_mitm()
         keypath = keypath.replace('h', '\'')
         keypath = keypath.replace('H', '\'')
 
-        if p2sh_p2wpkh:
-            addr_fmt = AF_P2WSH_P2SH if redeem_script else AF_P2WPKH_P2SH
-        elif bech32:
-            addr_fmt = AF_P2WSH if redeem_script else AF_P2WPKH
+        if addr_type == AddressType.SH_WIT:
+            addr_fmt = AF_P2WPKH_P2SH
+        elif addr_type == AddressType.WIT:
+            addr_fmt = AF_P2WPKH
+        elif addr_type == AddressType.LEGACY:
+            addr_fmt = AF_CLASSIC
         else:
-            addr_fmt = AF_P2SH if redeem_script else AF_CLASSIC
+            raise BadArgumentError("Unknown address type")
 
-        if redeem_script:
-            keypaths = keypath.split(',')
-            script = a2b_hex(redeem_script)
-
-            N = len(keypaths)
-
-            if not 1 <= N <= 15:
-                raise BadArgumentError("Must provide 1 to 15 keypaths to display a multisig address")
-
-            min_signers = script[0] - 80
-            if not 1 <= min_signers <= N:
-                raise BadArgumentError("Either the redeem script provided is invalid or the keypaths provided are insufficient")
-
-            if not script[-1] == 0xAE:
-                raise BadArgumentError("The redeem script provided is not a multisig. Only multisig scripts can be displayed.")
-
-            if not script[-2] == 80 + N:
-                raise BadArgumentError("Invalid redeem script, second last byte should encode N")
-
-            xfp_paths = []
-            for xfp in keypaths:
-                if '/' not in xfp:
-                    raise BadArgumentError('Invalid keypath. Needs a XFP/path: ' + xfp)
-                xfp, p = xfp.split('/', 1)
-
-                xfp_paths.append(str_to_int_path(xfp, p))
-
-            payload = CCProtocolPacker.show_p2sh_address(min_signers, xfp_paths, script, addr_fmt=addr_fmt)
-        # single-sig
-        else:
-            payload = CCProtocolPacker.show_address(keypath, addr_fmt=addr_fmt)
+        payload = CCProtocolPacker.show_address(keypath, addr_fmt=addr_fmt)
 
         address = self.device.send_recv(payload, timeout=None)
+        assert isinstance(address, str)
 
         if self.device.is_simulator:
             self.device.send_recv(CCProtocolPacker.sim_keypress(b'y'))
-        return {'address': address}
+        return address
 
-    # Setup a new device
-    def setup_device(self, label='', passphrase=''):
+    @coldcard_exception
+    def display_multisig_address(
+        self,
+        addr_type: AddressType,
+        multisig: MultisigDescriptor,
+    ) -> str:
+        if not multisig.is_sorted:
+            raise BadArgumentError("Coldcards only allow sortedmulti descriptors")
+
+        self.device.check_mitm()
+
+        if addr_type == AddressType.SH_WIT:
+            addr_fmt = AF_P2WSH_P2SH
+        elif addr_type == AddressType.WIT:
+            addr_fmt = AF_P2WSH
+        elif addr_type == AddressType.LEGACY:
+            addr_fmt = AF_P2SH
+        else:
+            raise BadArgumentError("Unknown address type")
+
+        if not 1 <= len(multisig.pubkeys) <= 15:
+            raise BadArgumentError("Must provide 1 to 15 keypaths to display a multisig address")
+
+        redeem_script = (80 + int(multisig.thresh)).to_bytes(1, byteorder="little")
+
+        if not 1 <= multisig.thresh <= len(multisig.pubkeys):
+            raise BadArgumentError("Either the redeem script provided is invalid or the keypaths provided are insufficient")
+
+        xfp_paths = []
+        sorted_keys = sorted(zip([p.get_pubkey_bytes(0) for p in multisig.pubkeys], multisig.pubkeys))
+        for pk, p in sorted_keys:
+            xfp_paths.append(p.get_full_derivation_int_list(0))
+            redeem_script += len(pk).to_bytes(1, byteorder="little") + pk
+
+        redeem_script += (80 + len(multisig.pubkeys)).to_bytes(1, byteorder="little")
+        redeem_script += b"\xae"
+
+        payload = CCProtocolPacker.show_p2sh_address(multisig.thresh, xfp_paths, redeem_script, addr_fmt=addr_fmt)
+
+        address = self.device.send_recv(payload, timeout=None)
+        assert isinstance(address, str)
+
+        if self.device.is_simulator:
+            self.device.send_recv(CCProtocolPacker.sim_keypress(b'y'))
+        return address
+
+    def setup_device(self, label: str = "", passphrase: str = "") -> bool:
+        """
+        The Coldcard does not support setup via software.
+
+        :raises UnavailableActionError: Always, this function is unavailable
+        """
         raise UnavailableActionError('The Coldcard does not support software setup')
 
-    # Wipe this device
-    def wipe_device(self):
+    def wipe_device(self) -> bool:
+        """
+        The Coldcard does not support wiping via software.
+
+        :raises UnavailableActionError: Always, this function is unavailable
+        """
         raise UnavailableActionError('The Coldcard does not support wiping via software')
 
-    # Restore device from mnemonic or xprv
-    def restore_device(self, label='', word_count=24):
+    def restore_device(self, label: str = "", word_count: int = 24) -> bool:
+        """
+        The Coldcard does not support restoring via software.
+
+        :raises UnavailableActionError: Always, this function is unavailable
+        """
         raise UnavailableActionError('The Coldcard does not support restoring via software')
 
-    # Begin backup process
     @coldcard_exception
-    def backup_device(self, label='', passphrase=''):
+    def backup_device(self, label: str = "", passphrase: str = "") -> bool:
+        """
+        Creates a backup file in the current working directory. This file is protected by the
+        passphrase shown on the Coldcard.
+
+        :param label: Value is ignored
+        :param passphrase: Value is ignored
+        """
         self.device.check_mitm()
 
         ok = self.device.send_recv(CCProtocolPacker.start_backup())
@@ -300,30 +357,41 @@ class ColdcardClient(HardwareWalletClient):
         result = self.device.download_file(result_len, result_sha, file_number=0)
         filename = time.strftime('backup-%Y%m%d-%H%M.7z')
         open(filename, 'wb').write(result)
-        return {'success': True, 'message': 'The backup has been written to {}'.format(filename)}
+        return True
 
-    # Close the device
-    def close(self):
+    def close(self) -> None:
         self.device.close()
 
-    # Prompt pin
-    def prompt_pin(self):
+    def prompt_pin(self) -> bool:
+        """
+        The Coldcard does not need a PIN sent from the host.
+
+        :raises UnavailableActionError: Always, this function is unavailable
+        """
         raise UnavailableActionError('The Coldcard does not need a PIN sent from the host')
 
-    # Send pin
-    def send_pin(self, pin):
+    def send_pin(self, pin: str) -> bool:
+        """
+        The Coldcard does not need a PIN sent from the host.
+
+        :raises UnavailableActionError: Always, this function is unavailable
+        """
         raise UnavailableActionError('The Coldcard does not need a PIN sent from the host')
 
-    # Toggle passphrase
-    def toggle_passphrase(self):
+    def toggle_passphrase(self) -> bool:
+        """
+        The Coldcard does not support toggling passphrase from the host.
+
+        :raises UnavailableActionError: Always, this function is unavailable
+        """
         raise UnavailableActionError('The Coldcard does not support toggling passphrase from the host')
 
-def enumerate(password=''):
+def enumerate(password: str = "") -> List[Dict[str, Any]]:
     results = []
     devices = hid.enumerate(COINKITE_VID, CKCC_PID)
     devices.append({'path': CC_SIMULATOR_SOCK.encode()})
     for d in devices:
-        d_data = {}
+        d_data: Dict[str, Any] = {}
 
         path = d['path'].decode()
         d_data['type'] = 'coldcard'
@@ -339,7 +407,7 @@ def enumerate(password=''):
         with handle_errors(common_err_msgs["enumerate"], d_data):
             try:
                 client = ColdcardClient(path)
-                d_data['fingerprint'] = client.get_master_fingerprint_hex()
+                d_data['fingerprint'] = client.get_master_fingerprint().hex()
             except RuntimeError as e:
                 # Skip the simulator if it's not there
                 if str(e) == 'Cannot connect to simulator. Is it running?':
